#! /usr/bin/perl

#                    Send Feedback from Web Page

#                          by John Walker
#                    Version 1.0  --  October 2004
#   	    	     Version 1.1  --  December 2006


    use strict;
    use warnings;
    use Fcntl ':flock';     	# Import flock() LOCK_ constants

#   Set $CGI_Directory to the CGI binary program directory on
#   your Web server.

    my $CGI_Directory = '/var/www/cgi-bin';

#   If $email is non-blank, feedback will be mailed to
#   the designated address using the $mailer program, which
#   is invoked with the specified $mailerpath.  $mailer is
#   used to specify the command line syntax and message protocol
#   used to send the mail, as opposed to the program to execute.

    my $email = 'feedback@ratburger.net';   # E-mail address to which you wish feedback sent

#   Typical settings for $mail and $mailerpath
#
#   Standard Unix mail program:
#
#   	$mailer     	Set to 'mail', or 'Mail', or 'mailx'--all are currently equivalent
#   	$mailerpath 	Linux:	    '/bin/mail' or '/usr/bin/Mail'
#   	    	    	Solaris:    '/usr/bin/mailx'
#   	    	    	IRIX	    '/usr/sbin/Mail'
#
#   Unix sendmail program:
#
#   	$mailer     	Set to 'sendmail'
#   	$mailerpath 	Linux:	    '/usr/lib/sendmail' or '/usr/sbin/sendmail.sendmail' or who knows?
#   	    	    	Solaris:    '/usr/lib/sendmail'
#   	    	    	IRIX	    '/usr/lib/sendmail'
    my $mailer = 'mail';
    my $mailerpath = '/usr/bin/Mail';

#   If $htmlMail is set nonzero, feedback will sent as HTML mail with
#   links to the page of origin, the submitter's E-mail address, and
#   to add/remove the user from the white list.  You can only enable
#   $htmlMail if the $mailer is 'sendmail'.

    my $htmlMail = 0;

#   If "logresolve" is installed on your system, specify the path below.
#   IP addresses will be resolved and the domain name displayed if
#   the name resolves.  If you don't have logresolve or prefer numeric
#   IP addresses, set this variable to the null string.

    my $logresolve = '/usr/bin/logresolve';

#   Define $spellCmd as the path to the command-line spelling checker
#   program on your system.  This program must be able to read from
#   standard input and write the list of suspect words to standard
#   output.  Set to the null string to disable spelling checking.

    my $spellCmd = '/usr/bin/spell';

#   Set $actionscript to the CGI script which should be run when
#   the user submits feedback.  This is usually this self-same program,
#   but you can set it to an argument dump program or other diagnostic
#   tool whilst debugging.

    my $actionscript = '/cgi-bin/FeedbackForm.pl';

#   Set $workingDirectory to the absolute directory path in
#   which the working files are kept.  This directory must
#   be readable and writable by the user under which this
#   script runs.

    my $workingDirectory = "$CGI_Directory/FeedbackForm";

    my $whiteList = 'whitelist.csv';	    # White list database
    my $blackList = 'blacklist.csv'; 	    # Black list database
    my $greenList = 'greenlist.csv';	    # Last solved problem database

#   If you'd like a log file to be written listing successsful and
#   unsuccessful feedback attempts, specify the absolute path name
#   of the log file here.  The directory must be one in which the
#   CGI user is permitted to create and write files.  Setting
#   $logFile to '' disables logging.

    my $logFile = "$workingDirectory/log.csv";

#   Lines longer than $wrapMax will be wrapped into multiple
#   lines no longer than $wrapMax (unless a single token is
#   longer).  We don't wrap multiple lines together, since the
#   assumption is that long lines result from bozo paragraph
#   per line data entry.  Setting $wrapMax to 0 disables wrapping.

    my $wrapMax = 64;	    	    	# Wrap lines longer than this

#   If $headerText is non-null, it will appear at the start
#   of the feedback form page.  This is HTML code, included
#   instead of the default feedback title.  This
#   string is evaluated and may contain Perl variable
#   interpolations.  Note that any double quote marks which
#   appear in this string must be escaped with "\", and that
#   the backslash must be preserved in the string value (which
#   means it's best to use single quotes to define $headerText).

    my $headerText = '';

#  If $footerText is non-null, it will be appended to the end of the
#  feedback form.  This is HTML code included right before the end of
#  the main division in the form. This string is evaluated
#  and may contain Perl variable interpolations.  Note that any
#  double quote marks which appear in this string must be escaped
#  with "\", and that the backslash must be preserved in the string
#  value (which means it's best to use single quotes to define
#  $footerText).

    my $footerText = '';

#   Define how long an IP address remains in the blacklist.  If you
#   set $blacklistTime to zero, the blacklist mechanism will be
#   entirely disabled.

    my ($seconds, $minutes, $hours, $days) = (1, 60, 60 * 60, 24 * 60 * 60);

    my $blacklistTime = 2 * $hours;

#   Define how long a problem remains available for solution until it
#   expires.  This can be used to limit the time the user can work
#   on the problem, but is also used to keep track of the time stamp
#   of the last problem solved from an IP address, which keeps
#   users from sending multiple feedbacks by going back to a previously
#   solved problem.  After $expirationTime has elapsed, we can purge
#   the record for the solved problem from the $greenList database.
#   If you set $expirationTime to zero, problems will never expire
#   and no $greenList will be maintained.

    my $expirationTime = 1 * $hours;

#   The confirmation page which appears after feedback is accepted
#   will return to the original referring page after $returnTime
#   seconds.  If $returnTime is zero, it won't automatically
#   return.

    my $returnTime = 10 * $seconds;

#   How many incorrect replies we'll tolerate before placing an submitter's
#   IP address on the blacklist.

    my $maxtries = 3;	    	    	# Maximum errors before ditching feedback

#   The following parameters control the mathematical problem the submitter
#   is asked to solve as a "sentience test"/

    my $sentience = 1;	    	    	# Nonzero to require sentience test
    my $minorder = 1;	    	    	# Minimum order of problem
    my $maxorder = 1;	    	    	# Maximum order of problem
    my $mincoeff = 1;    	    	# Minimum coefficient
    my $maxcoeff = 100;    	    	# Maximum coefficient

    my $minanswer = 4;   	    	# Minimum answer
    my $maxanswer = 100;    	    	# Maximum answer
    my $leastanswer = 2;    	    	# Force answers to >= this absolute value

#   The "Master Key" is used to obfuscate values passed
#   back and forth to the user.  You can generate your own
#   master keys with:
#   	http://www.fourmilab.ch/javascrypt/pass_phrase.html
#   check "Hexadecimal" and press "New Seed", then copy the hexadecimal
#   value and paste it into the next line.  If you can't be bothered to
#   to this, you can generate a master key of less certain pedigree by
#   invoking this program as "perl FeedbackForm.pl masterkey" and pasting the
#   resulting declaration into the line below.

    my $masterKey = '00';

#   If $restrictReferer is non-null, it will be used as a regular
#   expression used to test the HTTP_REFERER field we receive from
#   the Web server.  This can (and should) be used to restrict access
#   to this script to pages originating at the site where it's hosted.
#   If your site has several aliases (ratburger.org, ratburger.net,
#   etc.) or permits access by IP address or IP address range, you'll
#   have to craft a regular expression which matches everything you
#   may see as a referer.

    my $restrictReferer = '';

#   If $restrictIP is non-null, it will be used as a regular expression
#   to test the IP address from which white-list requests are sent.
#   If the IP address does not match, the request will be rejected.
#   Use this if you always manage your mail from a known IP address or
#   range.  If you process such requests from different IP addresses,
#   leave this null.

    my $restrictIP = '';

#   If $cribIP is non-null, IP addresses which match the regular expression
#   it's defined as will have the solution to the equation filled into
#   the text box in the feedback page.  This is handy when testing new
#   versions of the program as you don't need to actually solve the equation
#   every time in a long series of tests, and since you can still edit the
#   value in the box, it permits testing both right and wrong answers, which
#   whitelisting does not.

    my $cribIP =  '';

#   You can override the $email specification given above with a
#   hidden item in the form like the following:
#   	<input type="hidden" name="j" value="%D3%F6E%60y%14%0A%CE%EC%93%0C%0E%E6%7E%A8JEp">
#   To prevent the E-mail address from being "harvested" and buried
#   in a mountain of spam, it is encrypted using the $masterKey defined
#   above.  To encrypt an E-mail address, run this program from the command
#   line with:
#   	perl FeedbackForm.pl mailcode mymail@ratburger.net
#   which will print the "<input" declaration for the E-mail address
#   mymail@ratburger.net encrypted with your master key.  Paste this
#   into the form containing the Feedback button to override the $email
#   specification in this script.  DON'T DO THIS UNLESS YOU ABSOLUTELY HAVE TO!
#   An E-mail specification embedded within a script the user can't see
#   will always be more secure than one, however encoded, which appears in
#   a page the user can examine.  Overriding E-mail addresses is possible
#   only if $overrideEmail is set to 1.

    my $overrideEmail = 0; 	    	# THINK before you set this to 1!

#   If you must set $overrideEmail to 1, you should restrict E-mail overrides
#   to your own domain by forcing them to match the regular expression
#   $restrictEmail.  Otherwise, should your $masterKey be compromised
#   or somebody figure out how to break the encryption in some other
#   manner, your feedback page could be hijacked for use by other sites
#   and, even worse, used to send junk mail to third parties.

    my $restrictEmail = '';

#   You can specify the URL to which the "Back" link points by
#   including a hidden "return" field like the following:
#   	<input type="hidden" name="return" value="http://www.ratburger.net/whizzer/whizzer.html">
#   If you don't include a "return" item, the Back link will point
#   back to the page which invoked this script, courtesy of
#   the HTTP_REFERER environment item provided by the server.

#   You can specify the title of the referring page by
#   including a hidden "pagetitle" field like the following:
#   	<input type="hidden" name="pagetitle" value="Fox's UFO Bookshelf">
#   If you don't include a "pagetitle" item, the URL will be used as
#   the title.

#   You can specify the text of the link back to the parent
#   document (or whatever document you specified in the "return"
#   item) by including a hidden item of the form:
#   	<input type="hidden" name="backlink" value="Back to Fox's UFO Bookshelf">
#   If no "backlink" field appears in the form, the link will
#   bear the imaginative legend "Back to <URL>".

#   You shouldn't need to change anything below this line.

    my $prog = "FeedbackForm";
    my @worder = qw(trivial linear quadratic cubic quartic);
    my @strikewords = qw(Zero One Two Three Four Five Six Seven Eight Nine Ten);

    #	Test for "test" mode invocation to test configuration settings

    if (($#ARGV == 0) && ($ARGV[0] eq "test")) {
    	testConfiguration();
    }

    #	Decode Master Key into a byte vector

    my $mk = '';
    while ($masterKey =~ s/(\w\w)//) {
    	$mk .= pack("C", hex($1));
    }

    #	Test for special invocation with "mailcode" to encrypt an E-mail address
    #	with the master key.

    if (($#ARGV >= 0) && ($ARGV[0] eq 'mailcode')) {
    	encodeEmailAddress();
 	exit(0);
    }

    #	If invoked with an argument of "masterkey", generate a more or less
    #	pseudorandom master key declaration.

    if (($#ARGV == 0) && ($ARGV[0] eq "masterkey")) {
    	generateMasterKey();
    }

    parse_query_string();

    my $IPaddress = trim($ENV{'REMOTE_ADDR'});

    if (defined($ENV{'WWW_addwhite'})) {
    	addWhiteList();
    	exit(0);
    }
    if (defined($ENV{'WWW_remwhite'})) {
    	remWhiteList();
    	exit(0);
    }

    #	If $restrictReferer is defined, test the HTTP_REFERER environment
    #	variable for a match.  If it fails to match die, leaving only a
    #	message in the server error log as evidence.  The (ab)user who
    #	is trying to hijack the feedback form will see only a "500 server
    #	misconfiguration" error, and will have no idea why the access
    #	actually failed.

    if (($restrictReferer ne '') &&
    	((!defined($ENV{'HTTP_REFERER'})) || (!($ENV{'HTTP_REFERER'} =~ m/$restrictReferer/)))) {
    	die("$prog: Possible hijack attempt from $IPaddress, HTTP_REFERER=$ENV{'HTTP_REFERER'} failed to match\n");
    }

    #	We must initialise all of these to blanks to avoid warnings about
    #	use of uninitialised values when creating a blank feedback form from
    #	a minimalist request button.

    my ($referer, $subject, $name, $umail, $text, $backlink) = ('', '', '', '', '', '');

    #	When we're invoked from a minimalist page, we'll receive URL of the
    #	page which called us as HTTP_REFERER.  When invoked from a form we
    #	generated (or a page which explicitly specifies a referer URL with a
    #	"hidden" item, we take the parent page from WWW_referer.  The following
    #	code makes that happen.

    if (defined $ENV{'WWW_referer'}) {
    	$referer = trim($ENV{'WWW_referer'});
    } else {
	if (defined $ENV{'HTTP_REFERER'}) {
    	    $referer = trim($ENV{'HTTP_REFERER'});
	} else {
    	    $referer = '';
	}
    }

    $name = trim($ENV{'WWW_name'}) if (defined $ENV{'WWW_name'});
    $umail = trim($ENV{'WWW_email'}) if (defined $ENV{'WWW_email'});
    my $whiteListed = ($umail ne '') && inWhiteList($umail);
    $subject = $ENV{'WWW_subject'} if (defined $ENV{'WWW_subject'});
    $text = trim($ENV{'WWW_text'}) if (defined $ENV{'WWW_text'});
    my $send = trim($ENV{'WWW_send'});
    my $preview = trim($ENV{'WWW_preview'});
    my $spell = trim($ENV{'WWW_spell'});
    if (defined $ENV{'WWW_j'}) {
    	my $ec = unescape_URL($ENV{'WWW_j'});
	my $em = decrypt($ec);;
	if ($overrideEmail && (($restrictEmail eq '') || ($em =~ m/$restrictEmail/))) {
    	    $email = $em;
	} else {
print(STDERR "Override e-mail ($em) forbidden from $IPaddress $referer\n");
    	}
    }
    $referer = $ENV{'WWW_return'} if (defined $ENV{'WWW_return'});

    #	If the referring form includes a hidden "pagetitle" item,
    #	its value is used as the title for the page used as the
    #	link back to the referer.  If no "pagetitle" is specified,
    #	the URL is used as the title.

    my $pagetitle = $referer;
    if (defined $ENV{'WWW_pagetitle'}) {
    	$pagetitle = trim($ENV{'WWW_pagetitle'});
    }

    if (defined $ENV{'WWW_backlink'}) {
    	$backlink = $ENV{'WWW_backlink'};
    } else {
    	$backlink = "Back to $referer";
    }

    #	Parse "mode" field if present

    my ($testmode, $dumpmode);

    if (defined $ENV{'WWW_mode'}) {
    	$testmode = ($ENV{'WWW_mode'} =~ m/test/i);
	$dumpmode = ($ENV{'WWW_mode'} =~ m/dump/i);
    }

    #	Check the blacklist to see if this IP address has struck
    #	out.  If so, generate the strikeout document and leave it
    #	at that.

    my ($ibl, $blt) = inBlackList($IPaddress, 0);
    if ($ibl >= $maxtries) {
    	sendStrikeout($IPaddress, $blt, $referer, $backlink);
	addLog("Struckout", $IPaddress, $umail, $name, $subject, $referer);
	exit(0);
    }

    my ($goof, $pratfall) = ('', '');

    if ($send) {
    	my $eqsolved = '';

	if ($subject eq '') {
    	    $goof .= "<li>Please specify a Subject for your feedback message.</li>\n";
	}

	if ($text eq '') {
    	    $goof .= "<li>There's no message! Cat got your tongue?</li>\n";
	}
	if ($sentience && ($goof eq '')) {
    	    if (($umail eq '') || (!$whiteListed)) {
		#	Verify correct solution to equation
		my $e = $ENV{'WWW_e'};
		my $x = $ENV{'WWW_x'};
		my ($dok, $dxtime, $drhs, @dc) = decodeProblem($e);
		if ($dok) {
		    if (($expirationTime > 0) && (($dxtime + $expirationTime) < time())) {
		    	$goof .= "<li>Sorry, the time allowed to solve that equation expired\n";
			$goof .= "at " . edTime($dxtime + $expirationTime) . " Universal Time.\n";
			$goof .= "(Current Universal Time is " . edTime(time()) . ".)\n";
			$goof .= "Please solve the new equation given below.</li>\n";
		    	addLog("Expired", $IPaddress, $umail, $name, $subject, $referer);
			$ENV{'WWW_e'} = '';
		    } else {
		    	$x =~ s/\s//g;
			if ($x eq '') {
		    	    $goof .= "<li>You didn't enter a value for <i>x</i>.\n";
		    	    $goof .= "You must solve the equation to send feedback.</li>\n";
			    addLog("Nosol", $IPaddress, $umail, $name, $subject, $referer);
			} elsif (($x =~ m/[^\-\+0-9]/) || ($x eq '')) {
		    	    $goof .= "<li>Your value for <i>x</i> isn't even wrong!  &ldquo;" .
			    	htmlquote($ENV{'WWW_x'}) . "&rdquo; is not a number.</li>\n";
			    addLog("Notnum", $IPaddress, $umail, $name, $subject, $referer, $x);
			} else {
			    my ($sok, $urhs) = testSolution($drhs, $x, @dc);
			    if ($sok) {
			    	my $glt = inGreenList($IPaddress, 0);
				if ($dxtime <= $glt) {
				    $goof .= "<li>You've already solved that equation!  You only get to\n";
				    $goof .= "send one feedback message per equation you solve.  Please\n";
				    $goof .= "solve the new equation below.</li>\n";
				    addLog("Reused", $IPaddress, $umail, $name, $subject, $referer);
				    $ENV{'WWW_e'} = '';
				} else {
				    $eqsolved = surrenderEquation($drhs, @dc);
				    #   A correct solution purges the IP address from the blacklist.
				    inBlackList($IPaddress, -1);
				    #   A correct solution adds the time stamp of the last problem
				    #   solved to the greenlist.
				    inGreenList($IPaddress, $dxtime);
				}
			    } else {
		    		$pratfall .= "Sorry, your solution to the equation:\n";
				$pratfall .= "<br /> &nbsp; &nbsp; &nbsp; " . renderEquation($drhs, @dc) . " <br />\n";
				$pratfall .= "is incorrect.  With your value of <i>x</i> = $x, the right hand\n";
				$pratfall .= "side would be $urhs, not $drhs.\n";
				addLog("Goof", $IPaddress, $umail, $name, $subject, $referer, surrenderEquation($drhs, @dc), $drhs, $urhs, $x);
				$ENV{'WWW_e'} = '';
    	    		    }
			}
		    }
		} else {
		    $goof .= "<li>Nice try, but tampering with the problem isn't appreciated.</li>\n";
		    inBlackList($IPaddress, 1);
		    addLog("Tamper", $IPaddress, $umail, $name, $subject, $referer);
		    $ENV{'WWW_e'} = '';
		}
	    }
	}

    	#	If all was well, send the feedback via E-mail to the designated address
    	#	and send the user a confirmation that the feedback has been sent.

    	if (($goof eq '') && ($pratfall eq '')) {
	    addLog($testmode ? "Test" : "Sent", $IPaddress, $umail, $name, $subject, $referer, $eqsolved, $email);
	    if (!$testmode) {
    	    	sendMail($email, $name, $umail, $IPaddress, $referer, $subject, $text);
	    }
    	    sendConfirmation($email, $name, $umail, $IPaddress, $referer, $subject, $text);
    	    exit(0);
    	}
    }

    #	Generate feedback form for submitter.

    sendForm($goof, $pratfall, $email, $name, $umail, $IPaddress, $referer, $subject, $text);
    close(STDOUT);

    #	Send an E-mail message

    sub sendMail {
    	my ($email, $name, $umail, $IPaddress, $referer, $subject, $text) = @_;

    	my ($useMail, $useSendmail) = (0, 0);
	if (($email ne '') && ($mailer ne '')) {
	    my $dotty = 0;
	    my $m;
	    if (($mailer eq 'mailx') || ($mailer eq 'mail') || ($mailer eq 'Mail')) {
	    	$useMail = 1;
	    	$m = "$mailerpath -s '[Feedback] ($subject) from $referer'";
		$dotty = 1;
	    } elsif ($mailer eq 'sendmail') {
	    	$useSendmail = 1;
	    	$m = "$mailerpath";
		$dotty = 1;
    	    }

	    my $fqdn = '';
	    if ($logresolve ne '') {
		$fqdn = `echo $IPaddress | $logresolve`;
		chomp($fqdn);
	    }

	    #	If $wrapMax is nonzero, wrap long lines so they're no longer than $wrapMax

	    if ($wrapMax > 0) {
	    	my $t = $text;
		$text = '';
		if (!($t =~ m/\n$/)) {
		    $t .= "\n";
		}
		while ($t =~ s/^(.*\n)//) {
		    my $l = $1;
		    if (length($l) > $wrapMax) {
		    	$l = wrap($l, $wrapMax);
		    }
		    $text .= $l;
		}
	    }

	    #	Make absolutely sure there's an end of line sequence at the end of the text

	    if (!($text =~ m/\n$/)) {
		$text .= "\r\n";
	    }

	    #	If the selected mail transfer program cannot accept a period
	    #	on a line by itself, transform any such sequence in the
	    #	$text by appending a space before the period.

	    if ($dotty) {
    	    	$text =~ s/\r\n/\n/g;
	    	$text =~ s/^\.\n/\. \n/;
		$text =~ s/\n\.\n/\n\. \n/g;
	    }

	    #	Send message with configured mailer program

	    send_with_sendmail($m, $fqdn, $email, $name, $umail, $IPaddress, $referer, $subject, $text) if ($useSendmail);
	    send_with_mail($m, $fqdn, $email, $name, $umail, $IPaddress, $referer, $subject, $text) if ($useMail);
	}
    }

    #	Send mail with sendmail

    sub send_with_sendmail {
    	my ($m, $fqdn, $email, $name, $umail, $IPaddress, $referer, $subject, $text) = @_;

    	$text =~ s/\n/\r\n/g;

    	open(MAIL, "| $m $email") || die("Cannot create pipe to $m $email");

	print(MAIL "From FeedbackForm\r\n");
	print(MAIL "To: $email\r\n");
	print(MAIL "Subject: [Feedback] ($subject) from $referer\r\n");

   	if ($htmlMail) {
	    my $wlaction = '';
	    if ($umail ne '') {
	    	if (inWhiteList($umail)) {
		    $wlaction = " <a href=\"http://$ENV{\"SERVER_NAME\"}$actionscript?remwhite=y&amp;email=" .
			escape_URL(encrypt($umail)) . "\">Remove from white list</a>";
		} else {
		    $wlaction = " <a href=\"http://$ENV{\"SERVER_NAME\"}$actionscript?addwhite=y&amp;email=" .
			escape_URL(encrypt($umail)) . "\">Add to white list</a>";
		}
	    }
	    #	Construct a Subject: and Body: to plug into the
	    #	mailto: URL to initialise those fields for the reply.
	    my $esubject = $subject;
	    $esubject =~ s/^(Re:\s*)+//i;
	    $esubject = escape_URL("Re: $esubject");
#	    my $ebody = $text;
#	    $ebody =~ s/^/> /mg;
#	    $ebody =~ s/\r\n/\r\n\r\n/g;
#	    $ebody = escape_URL($ebody);
    	    my ($qtext, $qname, $qsubject, $qumail, $qpagetitle, $qipaddress, $qfqdn) =
	    	(htmlquote($text), htmlquote($name), htmlquote($subject),
		 htmlquote($umail), htmlquote($pagetitle),
		 htmlquote($IPaddress), htmlquote($fqdn));
	    print MAIL << "EOF";
Content-Type: text/html\r
\r
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"\r
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">\r
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">\r
<head>\r
<title>Feedback from: $qpagetitle</title>\r
<meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1" />\r
</head>\r
<body>\r
<b>From:</b> $qname &lt;<a href="mailto:$qumail?subject=$esubject">$qumail</a>&gt;$wlaction<br />\r
<b>Subject:</b> $qsubject<br />\r
<b>Feedback from:</b> <a href="$referer">$qpagetitle</a><br />\r
<b>Sent from:</b> $qipaddress  $qfqdn\r
\r
<pre width="$wrapMax" style="padding: 8px; background-color: rgb(192, 192, 224);">\r
$qtext\r
</pre>\r
</body>\r
</html>\r
EOF
	} else {
	    print(MAIL "\r\n");
	    generate_text_mail("\r", $m, $fqdn, $email, $name, $umail, $IPaddress, $referer, $subject, $text);
	}

    	print(MAIL ".\r\n");
    	close(MAIL);
    }

    #	Send mail with Unix mail program

    sub send_with_mail {
    	my ($m, $fqdn, $email, $name, $umail, $IPaddress, $referer, $subject, $text) = @_;

   	if ($htmlMail) {
	    die('Attempted to send $htmlmail with Unix mail program');
	}

    	open(MAIL, "| $m $email") || die("Cannot create pipe to $m $email");
    	generate_text_mail("", $m, $fqdn, $email, $name, $umail, $IPaddress, $referer, $subject, $text);
	close(MAIL);
    }

    #	Generate plain ASCII text mail

    sub generate_text_mail {
    	my ($cr, $m, $fqdn, $email, $name, $umail, $IPaddress, $referer, $subject, $text) = @_;

	print(MAIL "From: $name <$umail>$cr\n");
	print(MAIL "Subject: $subject$cr\n");
	print(MAIL "Feedback from page: $referer$cr\n");
	print(MAIL "Sent from: $IPaddress  $fqdn$cr\n");
	if ($umail ne '') {
	    if (inWhiteList($umail)) {
		print(MAIL "  Remove from white list?$cr\n    http://$ENV{\"SERVER_NAME\"}$actionscript?remwhite=y&email=" . escape_URL(encrypt($umail)));
	    } else {
	    	print(MAIL "  Add to white list:$cr\n    http://$ENV{\"SERVER_NAME\"}$actionscript?addwhite=y&email=" . escape_URL(encrypt($umail)));
	    }
	}
	print(MAIL "$cr\n");
	print(MAIL $text);
    }


    #	Send confirmation of feeback sent

    sub sendConfirmation {
    	my ($email, $name, $umail, $IPaddress, $referer, $subject, $text) = @_;

	$name = htmlquote($name);
	$umail = htmlquote($umail);
	$subject = htmlquote($subject);
	my $qpagetitle = htmlquote($pagetitle);

	my $rt = ($returnTime > 0) ? "<meta http-equiv=\"Refresh\" content=\"$returnTime; URL=$referer\" />" : '';

    	print << "EOF";
Content-type: text/html

<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
<head>
$rt
<title>Feedback sent</title>
<meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1" />
<style type="text/css">
    blockquote {
    	background-color: rgb(75%, 100%, 75%);
	color: #000000;
	padding: 10px;
	text-align: justify;
    }

    body {
    	background-color: #FFFFFF;
	color: #000000;
        margin-left: 10%;
        margin-right: 10%;
    }
</style>
<meta name="robots" content="noindex" />
</head>

<body>

<center>
EOF

    	if ($testmode) {
    	    print << "EOF";
<h2>Test mode.  Feedback not sent on page:<br /><a href="$referer">$qpagetitle</a></h2>
EOF
	} else {
    	    print << "EOF";
<h2>Feedback sent on page:<br /><a href="$referer">$qpagetitle</a></h2>
EOF
    	}

    	print << "EOF";
<h3>Thank you!</h3>

</center>

EOF

    	if ($whiteListed) {
    	print << "EOF";
<p />
<blockquote>
<h4>By the way, your E-mail address "$umail" is on our
white-list.  You needn't enter a solution to the equation
to send feedback.</h4>
</blockquote>
EOF
	}

	if ($returnTime > 0) {
    	    print << "EOF";
<p />
<b>
In $returnTime seconds your browser should automatically return you to the
page from which you sent feedback.  If it doesn't please click the link
below to return.
</b>
EOF
    	}

    	print << "EOF";

<h4><a href=\"$referer\">$backlink</a></h4>

</body>
</html>
EOF
    }

    #	Send a feedback request form, either blank or with values from
    #	the last attempt already filled in.

    sub sendForm {
    	my ($goof, $pratfall,, $email, $name, $umail, $IPaddress, $referer, $subject, $text) = @_;

	my $title = '';
	my ($qtext, $qname, $qumail, $qsubject, $qpagetitle) =
	    (htmlquote($text), htmlquote($name), htmlquote($umail), htmlquote($subject),
	     htmlquote($pagetitle));

	if ($headerText ne '') {
    	    $title = eval("\"$headerText\"");
	    if ($@) {
    	    	$goof .= "<li>Error in \$headerText expansion: $@<br /></li>\n";
	    }
	} else {
	    $title = << "EOF";
<center>
<h2>Send feedback on page:<br /><a href="$referer">$qpagetitle</a></h2>
</center>
EOF
	}

    	print << "EOF";
Content-type: text/html

<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
<head>
<title>Send feedback on $referer</title>
<meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1" />
<style type="text/css">
    body {
    	background-color: #FFFFFF;
	color: #000000;
        margin-left: 10%;
        margin-right: 10%;
    }

    blockquote.goof {
    	background-color: inherit;
    	color: #FF0000;
    }

    blockquote.goof blockquote {
    	font-weight: bold;
    }

    .preview {
    	padding: 8px;
	background-color: rgb(255, 255, 160);
    }
</style>
<meta name="robots" content="noindex" />
</head>

<body>

<form name="feedback" method="post" action="$actionscript">
<input type="hidden" name="referer" value="$referer" />

$title
EOF

    	#   Propagate form arguments we received to the child form

	propagate('j');     	    	# Encrypted feedback E-mail address
	propagate('mode');  	    	# Optional modes
	propagate('backlink');	    	# Back link description
	propagate('pagetitle');     	# Referring page title

	#   If there were forgivable entry errors, issue diagnostics for them

    	if ($goof ne '') {
	    print << "EOF";
<blockquote class="goof">
<h3>Regrettably, there are some quibbles with your feedback submission.
Please remedy them and try again.</h3>
<blockquote>
<ul>
$goof
</ul>
</blockquote>
</blockquote>
<p />
EOF
	}

	#   If the user entered a wrong answer or the equivalent (for
	#   example, tampering with the equation, call the strike and
	#   display the message.

    	my $strikeout = 0;
    	if ($pratfall ne '') {
	    my ($strikes, $blt) = inBlackList($IPaddress, 1);
	    my $sw = ($strikes < $#strikewords) ? $strikewords[$strikes] : $strikes;
	    $strikeout = ($strikes >= $maxtries);
	    my $yerout = $strikeout ? " &mdash; You're Out" : "";
	    print << "EOF";
<blockquote class="goof">
<h3>Strike $sw$yerout!</h3>
<blockquote>
$pratfall
</blockquote>
</blockquote>
<p />
EOF
	}

	#   If this was the last strike, so inform the batter.

	if ($strikeout) {
    	    	print << "EOF";
<b>Sorry, that's too many errors in a row.  Let's call it a day.</b>
<p />
EOF

	} else {

	    #	If a sentience test is configured, generate the problem to
	    #	pose to the user.  If there's a pending unsolved problem,
	    #	re-use it unless it's expired.

            if ($sentience) {
	    	my $dok = 0;
		my ($equation, $eprob, $answer, $wo);
	    	if (defined($ENV{'WWW_e'}) && ($ENV{'WWW_e'} ne '')) {
		    $eprob = $ENV{'WWW_e'};
		    my ($dxtime, $drhs, @dc);
		    ($dok, $dxtime, $drhs, @dc) = decodeProblem($eprob);
		    if ($dok) {
		    	if (($expirationTime > 0) && (($dxtime + $expirationTime) < time())) {
			    #	Existing problem has expired.  Force generation of a new one.
			    $dok = 0;
		    	} else {
		    	    $answer = '';
		    	    $equation = renderEquation($drhs, @dc);
			    if (defined($ENV{'WWW_x'})) {
				$answer = $ENV{'WWW_x'};
			    }
			    my $order = $#dc;
			    $wo = ($order <= $#worder) ? $worder[$order] : sprintf("%dth order", $order);
			}
		    }
		}

		#   There was no pre-existing equation (or it had expired or
		#   was corrupted).  Generate a new problem and clear any
		#   previously filled in answer.

		if (!$dok) {
            	    ($equation, $eprob, $answer, $wo) = poseProblem();
   	    	    if (!($IPaddress =~ m/$cribIP/)) {
			$answer = '';
		    }
		}

		my $sw = ($maxtries < $#strikewords) ? $strikewords[$maxtries] : $maxtries;
    	    	$sw =~ tr/A-Z/a-z/;

    	    	print << "EOF";
<h3>Sentience Test</h3>

<input type="hidden" name="e" value="$eprob" />

<p align="justify">
In order to avoid automatically-generated junk mail and input from
the innumerate, we ask you to take a moment and solve the following
$wo
equation.  Enter the correct value for <i>x</i> in the box.
EOF

    	    	if ($wo ne 'linear') {
    	    	    print << "EOF";
If the equation has more than one integer solution, any may be entered.
EOF
    	    	}

    	    	my $qanswer = htmlquote($answer);
    	    	print << "EOF";
If you enter an incorrect solution, you'll get
another chance but <em>with a new equation</em>.  Our patience
is not unlimited, however: $sw strikes and you're out.
</p>

<p align="center">
    <font size="+1">
    $equation
    <br />
    <i><u>x</u></i> = <input type="text" name="x" accesskey="x" value="$qanswer" size="6" maxlength="6" />
    </font>
</p>
EOF
    	    }

	    my $wm = ($wrapMax == 0) ? 70 : $wrapMax;

	    if ($testmode) {
    	    	print << "EOF";
<p align="center">
<b><font color="#FF0000">Test mode.  All input verification will work as
usual, but feedback will <em>not</em> actually be sent.</font></b>
</p>
EOF
	    }

    	    my $spellButton = ($spellCmd ne '') ?
	    	    '&nbsp; <input type="submit" value=" Spell Check " name="spell" accesskey="k" />' :
		    '';
    	    my $fwid = 60;  	    	# Field width for labels
    	    print << "EOF";
<center>
<table width="45%" border="border" cellpadding="5" bgcolor="#D0D0D0">

<tr>
<th align="left">
    The following fields are optional.  Anonymous comments are welcome
    but, of course, no reply will be possible.<br />
    <table>
    <tr>
    <th align="left" width="$fwid"><u>N</u>ame:</th>
    <td><input type="text" name="name" accesskey="n" value="$qname" size="60" maxlength="80" /><br /></td>
    </tr>
    <tr>
    <th align="left" width="$fwid"><u>E</u>-mail:</th>
    <td><input type="text" name="email" accesskey="e" value="$qumail" size="60" maxlength="80" /></td>
    </tr>
    </table>
</th>
</tr>
<tr>
<th align="left">
    <table>
    <tr>
    <th align="left" width="$fwid"><u>S</u>ubject:</th>
    <th align="left" width="$fwid"><input type="text" name="subject" accesskey="s" value="$qsubject" size="60" maxlength="80" /><br /></th>
    </tr>
    </table>
</th>
</tr>
<tr>
<th align="left">
<u>M</u>essage<br />
    <textarea name="text" accesskey="m" cols="$wm" rows="16">$qtext</textarea>
</th>
</tr>
<tr>
<td align="center">
<input type="submit" value=" Send " name="send" accesskey="d" /> &nbsp;
<input type="reset" value=" Clear " accesskey="c" /> &nbsp;
<input type="submit" value=" Preview " name="preview" accesskey="p" />
$spellButton
</td>
</tr>
</table>
</center>

EOF

    	    if ($preview) {
		my $pt = $text;
		my $wrapw = '';
		if ($wrapMax > 0) {
	    	    my $t = $pt;
		    $pt = '';
		    if (!($t =~ m/\n$/)) {
			$t .= "\n";
		    }
		    while ($t =~ s/^(.*\n)//) {
			my $l = $1;
			if (length($l) > $wrapMax) {
		    	    $l = wrap($l, $wrapMax);
			}
			$pt .= $l;
		    }
		    $wrapw = "width=\"$wrapMax\"";
		}
		$pt = htmlquote($pt);
		if ($wrapw eq '') {
		    $pt =~ s/\n/<p \/>\n/g;
    	    	    print << "EOF";
<h3>Preview</h3>
<blockquote class="preview">
$pt
</blockquote>
EOF
		} else {
    	    	    print << "EOF";
<h3>Preview</h3>
<pre $wrapw class="preview">
$pt
</pre>
EOF
    	    	    }
    	    }

    	    if ($spell && ($spellCmd ne '')) {
    	    	my $sfn = "$workingDirectory/spell$$.tmp";
    	    	if (open(SP, "| $spellCmd >$sfn")) {
		    print(SP $text . "\n");
		    close(SP);
		    open(SF, "<$sfn") ||
		    	die("Cannot reopen spelling file $workingDirectory/spell$$.tmp");
    	    	    my $pt = '';
		    while (<SF>) {
		    	$pt .= $_;
    	    	    }
		    close(SF);
		    unlink($sfn);
		    $pt =~ s/^\s+//;
		    $pt =~ s/\s+$//;
		    $pt =~ s/\s+/ /g;
		    if ($pt eq '') {
    	    	    	print << "EOF";
<h3 style="color: rgb(0,192,0);">No Misspelled Words</h3>
EOF
		    } else {
		    	$pt = wrap($pt, 72);
    	    	    	print << "EOF";
<h3>Possibly Misspelled Words</h3>
<pre style="padding: 8px; background-color: rgb(255, 160, 160);">
$pt
</pre>
EOF
    	    	    }
    	    	}
    	    }

	    print << "EOF";
<p />
<hr />
EOF

    	}

	#	Generate back link to originating page.  Note that the
	#	$referer URL is already quoted, encoding any impermissible
	#	characters as %XX escapes.

	if ($referer ne '') {
    	    print("<h4><a href=\"$referer\">$backlink</a></h4>");
	}

	my $ft = '';
	if ($footerText ne '') {
    	    $ft = eval("\"$footerText\"");
	}

	print << "EOF";

</form>
$ft
EOF

    	if ($dumpmode) {
    	    dump_environment();
	}

    	print << "EOF";
</body>
</html>
EOF
    }

    #	Generate the response when an IP address has struck out

    sub sendStrikeout {
    	my ($IPaddress, $blt, $referer, $backlink) = @_;

	my $sw = ($maxtries < $#strikewords) ? $strikewords[$maxtries] : $maxtries;
	my $et = edTime($blt + $blacklistTime);
	my $ct = edTime(time());

    	print << "EOF";
Content-type: text/html

<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
<head>
<title>Strike $sw -- You're Out!</title>
<meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1" />
<meta name="robots" content="noindex" />
</head>

<body bgcolor="#FFFFFF">

<center>
<h1><font color="#FF0000">Strike $sw &mdash; You're Out!</font></h1>
</center>

<blockquote>
<h3>
You've already struck out!
</h3>

<p />
You can try the feedback page again
at $et Universal (Greenwich Mean) Time.

<p />
(Current Universal time is $ct.)
</blockquote>
EOF

	#	Generate back link to originating page.  Note that the
	#	$referer URL is already quoted, encoding any impermissible
	#	characters as %XX escapes.

	if ($referer ne '') {
    	    print("<h4><a href=\"$referer\">$backlink</a></h4>");
	}

    	print << "EOF";
</body>
</html>
EOF
    }

    #	Add an E-mail address to the white-list

    sub addWhiteList {
    	my $email = decrypt($ENV{'WWW_email'});
	my $found = 0;
	my $what;

	if (($restrictIP ne '') && (!($IPaddress =~ m/$restrictIP/))) {
	    $what = "<h1>White list removal not permitted from IP address $IPaddress</h1>";
	} else {
	    if ((-d $workingDirectory) && (-f "$workingDirectory/$whiteList")) {
		$email = trim($email);
		open(WL, "+<$workingDirectory/$whiteList") ||
	    	    open(WL, "+<$workingDirectory/$whiteList")  ||
		    die("Cannot open or create $workingDirectory/$whiteList");
		flock(WL, LOCK_EX);
		seek(WL, 0, 0);
		my $l;
		while ($l = <WL>) {
	    	    $l = trim($l);
		    if ($email eq $l) {
			$found = 1;
			last;
		    }
		}
		if (!$found) {
	    	    seek(WL, 0, 2);
		    print(WL "$email\n");
		}
		flock(WL, LOCK_UN);
		close(WL);
	    }

    	    addLog("Addwhite", $IPaddress, $email);

 	    $email = htmlquote($email);
	    $what = "<h1>Added $email to White-List</h1>";
	}
   	print << "EOF";
Content-type: text/html

<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
<head>
<title>White-List Addition Request for $email</title>
<meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1" />
<meta name="robots" content="noindex" />
</head>

<body bgcolor="#FFFFFF">

<center>
$what
</center>

</body>
</html>
EOF
    }

    #	Remove an E-mail address from the white-list

    sub remWhiteList {
    	my $email = decrypt($ENV{'WWW_email'});
	my $changed = 0;
	my @wl;
	my $what;

	if (($restrictIP ne '') && (!($IPaddress =~ m/$restrictIP/))) {
	    $what = "<h1>White list removal not permitted from IP address $IPaddress</h1>";
	} else {
	    if ((-d $workingDirectory) && (-f "$workingDirectory/$whiteList")) {
		$email = trim($email);
		open(WL, "+<$workingDirectory/$whiteList") ||
	    	    open(WL, "+>$workingDirectory/$whiteList")  ||
		    die("Cannot open or create $workingDirectory/$whiteList");
		flock(WL, LOCK_EX);
		seek(WL, 0, 0);
		my $l;
		while ($l = <WL>) {
	    	    my $al = trim($l);
		    if ($email eq $al) {
			$changed = 1;
		    } else {
			push(@wl, $l);
		    }
		}
		if ($changed) {
	    	    seek(WL, 0, 0);
		    truncate(WL, 0);
		    print(WL @wl);
		}
		flock(WL, LOCK_UN);
		close(WL);
	    }

	    $email = htmlquote($email);
	    if ($changed) {
		$what = "<h1>Removed $email from White-List</h1>";
    	    	addLog("Delwhite", $IPaddress, $email);
	    } else {
		$what = "<h1>$email was not in the White-List!</h1>";
	    }
	}
    	print << "EOF";
Content-type: text/html

<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
    "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="en" lang="en">
<head>
<title>White-List Removal Request for $email</title>
<meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1" />
<meta name="robots" content="noindex" />
</head>

<body bgcolor="#FFFFFF">

<center>
$what
</center>

</body>
</html>
EOF
    }

    #	Propagate a form variable we received in the environment as a
    #	hidden form input in an HTML document we emit.

    sub propagate {
    	my ($fld) = @_;

	if (defined $ENV{"WWW_$fld"}) {
    	    my $v = htmlquote($ENV{"WWW_$fld"});
    	    print("<input type=\"hidden\" name=\"$fld\" value=\"$v\" />\n");
	}

    }

    #	Generate a mathematical problem to screen the submitter

    sub poseProblem {
	my $order = ($minorder + int(rand(1 + ($maxorder - $minorder))));
	my @coeff;
	for (my $p = $order; $p >= 0; $p--) {
	    $coeff[$p] = $mincoeff + int(rand(($maxcoeff - $mincoeff) + 1));
	    if ($coeff[$p] == 0) {
	    	$coeff[$p] = 1;
	    }
	}

	my $answer = $minanswer + int(rand(($maxanswer - $minanswer) + 1));
	if (abs($answer) < $leastanswer) {
	    $answer = $leastanswer * ($answer >= 0 ? 1 : -1);
	}
	my $rhs = 0;
	for (my $p = 0; $p <= $order; $p++) {
	    $rhs += ($answer ** $p) * $coeff[$p];
    	}

    	my $equation = renderEquation($rhs, @coeff);

	my $op1 = $order + 1;
	my $pprob = pack("C l$op1 l l", $order, @coeff, $rhs, time());
	my $csum = unpack("%8C*", $pprob);
	$pprob = pack("C", $csum) . $pprob;
    	$pprob = encrypt($pprob);
	$pprob = pack("u", $pprob);
	chomp($pprob);
	$pprob =  escape_ISO($pprob);
	my $wo = ($order <= $#worder) ? $worder[$order] : sprintf("%dth order", $order);
	return ($equation, $pprob, $answer, $wo);
    }

    #	Decode problem from encoded string

    sub decodeProblem {
    	my ($ps) = @_;

	my ($xcs, $xo, $xop1, @xc, $xrhs, $xtime);
	my $ude = unescape_ISO($ps);
	my $bps = unpack("u", $ude);
    	$bps = decrypt($bps);
    	($xcs, $xop1) = unpack("CC", $bps);
	my $csum = unpack("%8C*", substr($bps, 1));
	$xop1 += 3;
	($xcs, $xo, @xc) = unpack("C C l$xop1", $bps);
	$xtime = pop(@xc);
    	$xrhs = pop(@xc);
    	return ($csum == $xcs, $xtime, $xrhs, @xc);
    }

    #	Render an equation as HTML

    sub renderEquation {
    	my ($rhs, @coeff) = @_;

    	my @powers = ( '', '<i>x</i>', '<i>x</i>', '<i>x</i>' );

	my $order = $#coeff;

	my $equation;
	for (my $p = $order; $p >= 0; $p--) {
	    if (($coeff[$p] != 1) || ($p == 0)) {
	    	$equation .= $coeff[$p];
	    }
	    my $pterm;
	    if ($p > $#powers) {
		$pterm = "<i>x</i><sup>$p</sup>";
	    } else {
	    	$pterm = $powers[$p];
	    }
	    $equation .= $pterm;
	    if ($p > 0) {
    	    	$equation .= ' + ';
	    }
	}

	$equation .= " = $rhs";

	$equation =~ s/\+\s\-/\- /g;

	return $equation;
    }

    #	Strip HTML tags, leaving rendered equation as plain text

    sub surrenderEquation {
    	my $eq = renderEquation(@_);

	$eq =~ s:</?i>::g;
	$eq =~ s:<sup>(\d+)</sup>:^$1:g;
	$eq =~ s/\s+//g;
	return $eq;
    }

    #	Test if solution to equation is correct

    sub testSolution {
    	my ($rhs, $answer, @coeff) = @_;

	my $urhs = 0;
	for (my $p = 0; $p <= $#coeff; $p++) {
	    $urhs += ($answer ** $p) * $coeff[$p];
    	}

	return ($rhs == $urhs, $urhs);
    }

    #	Test whether a user's E-mail address is in the white list

    sub inWhiteList {
    	my ($email) = @_;
	my $found = 0;

	if ((-d $workingDirectory) && (-f "$workingDirectory/$whiteList")) {
	    $email = trim($email);
	    open(WL, "<$workingDirectory/$whiteList") || die("Cannot open $workingDirectory/$whiteList");
	    flock(WL, LOCK_EX);
	    seek(WL, 0, 0);
	    my $l;
	    while ($l = <WL>) {
	    	$l = trim($l);
		if ($email eq $l) {
		    $found = 1;
		    last;
		}
	    }
	    flock(WL, LOCK_UN);
	    close(WL);
	}
	return $found;
    }

    #	Test whether an IP address is in the blacklist

    sub inBlackList {
    	my ($ipaddr, $add) = @_;

	if ($blacklistTime == 0) {
	    return 0;
	}

	$ipaddr = trim($ipaddr);
	my @blacklist;
	my $bl;
	my $count = 0;
	my $changed = 0;
	my $remove = $add == -1;
	my $bltime = 0;

	if ($remove) {
	    $add = 0;
	}

	#   If the blacklist file exists, bring it into memory

	my $blo = 0;
	if ((-d $workingDirectory) && (-f "$workingDirectory/$blackList")) {
	    open($bl, "+<$workingDirectory/$blackList") || die("Cannot open $workingDirectory/$blackList");
	    flock($bl, LOCK_EX);
	    seek($bl, 0, 0);
	    @blacklist = <$bl>;
	    $blo = 1;
	}

	#   Remove expired items and any IP address to be removed from the blacklist

	for (my $i = 0; $i <= $#blacklist;) {
	    $blacklist[$i] =~ m/(^[0-9\.]+)\,(\d+)\,(\d+)/;
	    if (time() > ($3 + $blacklistTime)) {
    	    	splice(@blacklist, $i, 1);
		$changed = 1;
	    } elsif ($remove && ($ipaddr eq $1)) {
    	    	splice(@blacklist, $i, 1);
		$changed = 1;
	    } else {
	    	$i++;
	    }
	}

	#   Test whether the IP address is in the blacklist

	for (my $i = 0; $i <= $#blacklist; $i++) {
	    $blacklist[$i] =~ m/(^[0-9\.]+)\,(\d+)\,(\d+)/;
	    if ($1 eq $ipaddr) {
    	    	$bltime = $3;
    	    	$count = $2;
    	    	if ($add) {
		    $count++;
		    #	Increment count and update time in blacklist
		    $blacklist[$i] = "$1,$count," . ($bltime = time()) . "\n";
		    $changed = 1;
		}
	    }
	}

	#   If IP address wasn't already in the blacklist, add
	#   it with a count of 1 if requested by $add.

	if ($add && ($count == 0)) {
	    $count = 1;
	    push(@blacklist, "$ipaddr,$count," . time() . "\n");
    	    $changed = 1;
	}

	if ($changed) {

	    #   If the blacklist doesn't exist, create it

	    if (!$blo) {
		open($bl, "+>$workingDirectory/$blackList") || die("Cannot create $workingDirectory/$blackList");
		flock($bl, LOCK_EX);
    		$blo = 1;
	    }
	    seek($bl, 0, 0);
	    truncate($bl, 0);

	    #   Write the updated blacklist to the file

	    print($bl @blacklist);
	}

	if ($blo) {
	    flock($bl, LOCK_UN);
	    close($bl);
	}
	return ($count, $bltime);
    }

    #	Test whether an IP address is in the greenlist, or add
    #	an IP address and time stamp of the last good answer to
    #	the greenlist.

    sub inGreenList {
    	my ($ipaddr, $time) = @_;

	if ($expirationTime == 0) {
	    return 0;
	}

	$ipaddr = trim($ipaddr);
	my @greenlist;
	my $gl;
	my $changed = 0;
	my $gltime = 0;

	#   If the greenlist file exists, bring it into memory

	my $glo = 0;
	if ((-d $workingDirectory) && (-f "$workingDirectory/$greenList")) {
	    open($gl, "+<$workingDirectory/$greenList") || die("Cannot open $workingDirectory/$greenList");
	    flock($gl, LOCK_EX);
	    seek($gl, 0, 0);
	    @greenlist = <$gl>;
	    $glo = 1;
	}

	#   Remove expired items from the greenlist

	for (my $i = 0; $i <= $#greenlist;) {
	    $greenlist[$i] =~ m/(^[0-9\.]+)\,(\d+)/;
	    if (time() > ($2 + $expirationTime)) {
    	    	splice(@greenlist, $i, 1);
		$changed = 1;
	    } else {
	    	$i++;
	    }
	}

	#   Test whether the IP address is in the greenlist

	for (my $i = 0; $i <= $#greenlist; $i++) {
	    $greenlist[$i] =~ m/(^[0-9\.]+)\,(\d+)/;
	    if ($1 eq $ipaddr) {
    	    	$gltime = $2;
    	    	if ($time != 0) {
		    #	Update time in greenlist if requested
		    $gltime = $time;
		    $greenlist[$i] = "$1,$gltime\n";
		    $changed = 1;
		}
	    }
	}

	#   If IP address wasn't already in the greenlist, add
	#   it if a nonzero time argument was specified.

	if (($time != 0) && ($gltime == 0)) {
	    $gltime = $time;
	    push(@greenlist, "$ipaddr,$gltime\n");
    	    $changed = 1;
	}

	if ($changed) {

	    #   If the greenlist doesn't exist, create it

	    if (!$glo) {
		open($gl, "+>$workingDirectory/$greenList") || die("Cannot create $workingDirectory/$greenList");
		flock($gl, LOCK_EX);
    		$glo = 1;
	    }
	    seek($gl, 0, 0);
	    truncate($gl, 0);

	    #   Write the updated greenlist to the file

	    print($gl @greenlist);
	}

	if ($glo) {
	    flock($gl, LOCK_UN);
	    close($gl);
	}
	return $gltime;
    }

    #	Append an item to the log file

    sub addLog {
    	if ($logFile ne '') {
    	    my ($type, $ip, @args) = @_;

	    my @t = localtime(time());
	    my $time = sprintf("%d-%02d-%02d:%02d:%02d:%02d", $t[5] + 1900, $t[4] + 1, $t[3], $t[2], $t[1], $t[0]);
	    my $fqdn = '';
	    if ($logresolve ne '') {
		$fqdn = `echo $ip | $logresolve`;
	    }

	    my $l = "$type,$time," . quoteCSV($ip) . "," . quoteCSV($fqdn);
	    for (my $i = 0; $i <= $#args; $i++) {
	    	$l .= ',' . quoteCSV($args[$i]);
	    }
	    open(LF, "+>>$logFile") || die("Cannot open log file $logFile");
	    flock(LF, LOCK_EX);
	    seek(LF, 0, 2); 	    # May have grown between open and flock
	    print(LF "$l\n");
	    flock(LF, LOCK_UN);
	    close(LF);

	}
    }

    #	Quote a string for use in a CSV field

    sub quoteCSV {
    	my ($s) = @_;

	$s =~ s/[\000-\037\177-\241]//g;    	# Remove potentially damaging embedded characters
	if ($s =~ m/[\,\"]/) {
	    $s =~ s/"/""/g;
	    $s = '"' . $s . '"';
	}
	return $s;
    }

    #	Edit time and date

    sub edTime {
    	my ($utime) = @_;

	my @t = gmtime($utime);
	my @months = qw(January February March April May June July August September October November December);
	my @days = qw(Sunday Monday Tuesday Wednesday Thursday Friday Saturday);

	return sprintf("%02d:%02d on %s, %s %d, %d", $t[2], $t[1],
	    $days[$t[6]], $months[$t[4]], $t[3], $t[5] + 1900);
    }

    #	Dump a byte vector in hexadecimal

    sub hexdump {
    	my ($v) = @_;

	my $s = '';
	for (my $i = 0; $i < length($v); $i++) {
	    $s .= sprintf("%02X ", vec($v, $i, 8));
	}
	$s =~ s/\s$//;
	return $s;
    }

    #	Encrypt a byte vector with the master key

    sub encrypt {
    	my ($v) = @_;

	#   Prefix vector with initial vector
	$v = pack("C4", int(rand(256)), int(rand(256)), int(rand(256)), int(rand(256))) . $v;
	for (my $i = 1; $i < length($v); $i++) {
	    vec($v, $i, 8) ^= vec($v, $i - 1, 8) ^ vec($mk, $i % length($mk), 8);
	}

	return $v;
    }

    #	Decrypt a byte vector with the master key

    sub decrypt {
    	my ($v) = @_;

	for (my $i = length($v) - 1; $i > 0 ; $i--) {
	    vec($v, $i, 8) ^= vec($v, $i - 1, 8) ^ vec($mk, $i % length($mk), 8);
	}

	return substr($v, 4);	    # Return, less initial vector
    }

    #	Trim string by removing leading and trailing white space

    sub trim {
    	my ($s) = @_;

	if (!defined($s)) {
	    $s = '';
	}
	$s =~ s/^\s+//;     	# Delete leading blanks
	$s =~ s/\s+$//;     	# Delete trailing blanks
	return $s;
    }

    #	Quote user-supplied text for inclusion in log file

    sub iquote {
    	my ($s) = @_;

	$s =~ s/"/""/g;     	# Force quote marks with double quotes
	return "\"$s\"";    	# Return quoted string
    }

    #	Quote string for inclusion in an HTML document.  We quote
    #	'"' characters because the result may be used as the
    #	value of a tag attribute.

    sub htmlquote {
    	my ($s) = @_;

	$s =~ s/&/&amp;/g;
	$s =~ s/</&lt;/g;
	$s =~ s/>/&gt;/g;
	$s =~ s/"/&quot;/g;
	return $s;
    }

    #	Wrap a long text string.  Based upon Text::Wrap by
    #	David Muir Sharnoff, much simplified for use here.

    sub wrap {
	my ($t, $columns) = @_;

    	my ($ip, $xp) = ('', '');
	my $break = '\s';
	my $huge = 'overflow';	# 'wrap', 'die', or 'overflow'
	my $separator = "\n";

	my $r = "";
	my $tail = $t;
	my $lead = $ip;
	my $ll = $columns - length($ip) - 1;
	my $nll = $columns - length($xp) - 1;
	my $nl = "";
	my $remainder = "";

	pos($t) = 0;
	while ($t !~ /\G\s*\Z/gc) {
	    if ($t =~ /\G([^\n]{0,$ll})($break|\z)/xmgc) {
	    	$r .= $nl . $lead . $1;
	    	$remainder = $2;
	    } elsif ($huge eq 'wrap' && $t =~ /\G([^\n]{$ll})/gc) {
	    	$r .= $nl . $lead . $1;
	    	$remainder = $separator;
	    } elsif ($huge eq 'overflow' && $t =~ /\G([^\n]*?)($break|\z)/xmgc) {
	    	$r .= $nl . $lead . $1;
	    	$remainder = $2;
	    } elsif ($huge eq 'die') {
	    	die("Could not wrap '$t'");
	    } else {
		die("wrap: Invalid \$huge specification '$huge'");
	    }

	    $lead = $xp;
	    $ll = $nll;
	    $nl = $separator;
	}
	$r .= $remainder;

	$r .= $lead . substr($t, pos($t), length($t)-pos($t))
		if pos($t) ne length($t);

	return $r;
    }

    #	Dump the environment to standard output

    sub dumpenv {
    	my $k;

	for $k (sort keys %ENV) {
	    print("    $k = $ENV{$k}\n");
	}
    }

    #	Dump the environment as HTML

    sub dump_environment {
    	print << "EOF";
<center>
<h3>Program Environment</h3>
</center>
<pre>
EOF

    	dumpenv();

    	print << "EOF";
</pre>
EOF
    }

    #	Parse the CGI QUERY_STRING argument (placed in the environment when
    #	the Web server launches this script).  This function sets environment
    #	variables named WWW_varname for each field in the form with value the
    #	unquoted contents of each field.  This rather odd way of returning
    #	results ("Why not just use a regular hash?") is used to maintain
    #	compatibility with the UNCGI program.  A Perl script which previously
    #	required execution under the control of UNCGI may be run stand-alone
    #	by calling this function at the top.  This halves the number of processes
    #	the server must fork in order to process the request, and is one
    #	fewer thing for the beleaguered system administrator to install.

    sub parse_query_string {
    	my ($rmeth, @args, $instream, $var, $varname, $value);

	$rmeth = $ENV{'REQUEST_METHOD'};    # Request method ("GET" or "POST")
	$rmeth =~ tr/a-z/A-Z/;	    	    # Just in case server doesn't shout
	if ($rmeth eq 'GET') {
	    @args = split(/&/, $ENV{'QUERY_STRING'});
	} elsif ($rmeth eq 'POST') {
	    read(STDIN, $instream, $ENV{'CONTENT_LENGTH'});
	    @args = split(/&/, $instream);
	}
	foreach $var (@args) {
	    ($varname, $value) = split(/=/, $var);
	    $varname = unescape_URL($varname);
	    $value = unescape_URL($value);
	    #   The following code provides protection against server-side includes being
	    #   embedded in CGI arguments.  No sane site runs server-side includes, in my
	    #   opinion.  If you require this protection, uncomment these lines, but note
	    #   that it's (remotely) possible that binary encoded arguments may be corrupted
	    #   by this filter.
	    #	    #	Elide any server-side includes a mischevious user may have
	    #	    #	embedded in a value string.  This protects against mayhem if
	    #	    #	the value is included in an HTML reply.
	    #	    $value =~ s/<!--(.|\n)*-->//g;

	    #	Define an "uncgi" compatible environment variable for
	    #	each name/value pair.

	    $ENV{"WWW_$varname"} = $value;
	}
    }

    #	Escape a URL

    sub escape_URL {
    	my ($s) = @_;

    	$s =~ s/([^A-Za-z0-9\$\-_\.\!\*\(\)])/sprintf("%%%02X", ord($1))/eg;
	return $s;
    }

    #	Unescape a URL-encoded string

    sub unescape_URL {
    	my ($s) = @_;

	$s =~ tr/+/ /;
	$s =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;

	return $s;
    }

    #	Escape an ISO string

    sub escape_ISO {
    	my ($s) = @_;

    	$s =~ s/([^`~@\[\]\^_:;=\?A-Za-z0-9\$\-_\.\!\*'\(\)\xA1-\xFF])/sprintf("%%%02X", ord($1))/eg;
	return $s;
    }

    #	Unescape an ISO string

    sub unescape_ISO {
    	my ($s) = @_;

	$s =~ tr/+/ /;
	$s =~ s/%([a-fA-F0-9][a-fA-F0-9])/pack("C", hex($1))/eg;

	return $s;
    }

    #	Test the configuration and system environment

    sub testConfiguration {
	my $errors = 0;
	my $warnings = 0;

    	#   Test CGI directory
	if (!(-d $CGI_Directory)) {
	    print(STDERR '** $CGI_Directory' . " $CGI_Directory does not exist.\n");
	    $errors++;
	}

	#   Test working directory
	if (-d $workingDirectory) {
	    if (!(-w $workingDirectory)) {
	    	print(STDERR 'Warning: $workingDirectory' . " $workingDirectory is not writable by\n");
		print(STDERR "         user ID $< (" . getlogin() . ").  Be sure it is writable by the\n");
		print(STDERR "         user ID under which your Web server runs CGI programs.\n");
		$warnings++;
	    }
	} else {
	    print(STDERR '** $workingDirectory' . " $workingDirectory does not exist.\n");
	    $errors++;
	}

	#   Test for known mailer type

	if (($mailer eq 'mailx') || ($mailer eq 'mail') || ($mailer eq 'Mail') ||
	    ($mailer eq 'sendmail')) {
	    	if ($htmlMail && ($mailer ne 'sendmail')) {
	    	    print(STDERR '** $htmlMail set but $mailer set ' . "to \"$mailer\", not \"sendmail\".\n");
	    	    $errors++;
		}
	} else {
	    print(STDERR '** $mailer ' . "\"$mailer\" type unknown.\n");
	    $errors++;
    	}


	#   Test path to mailer program
	if ((-f $mailerpath)) {
	    if (!(-x $mailerpath)) {
	    	print(STDERR '** $mailerpath' . " $mailerpath exists but is not executable.\n");
	    	$errors++;
	    }
	} else {
	    print(STDERR '** $mailerpath' . " $mailerpath does not exist.\n");
	    $errors++;
	}

	#   Test path to logresolve and its functionality by
	#   trying a lookup of a.root-servers.net.
	if ((-f $logresolve)) {
	    if (-x $logresolve) {
	    	my $tip = '198.41.0.4';
	    	my $d = `echo $tip | $logresolve`;
		if (!($d =~ m/[A-Za-z]/)) {
		    chomp($d);
		    print(STDERR '** $logresolve' . " $logresolve does not seem to work.\n");
		    print(STDERR "   Resolve of '$tip' returned '$d', not a domain name.\n");
		    $errors++;
		}
	    } else {
	    	print(STDERR '** $logresolve' . " $logresolve exists but is not executable.\n");
	    	$errors++;
	    }
	} else {
	    print(STDERR '** $logresolve' . " $logresolve does not exist.\n");
	    $errors++;
	}

    	#   Verify that a valid master key has been specified
	if ($masterKey eq '00') {
	    print(STDERR '** $masterKey set to default.  Please set a $masterKey for encryption.' . "\n");
	    print(STDERR '   You can run this program with an argument of "masterkey" to generate' . "\n");
	    print(STDERR "   a new master key declaration.\n");
	    $errors++;
	} else {
	    if (((length($masterKey) & 1) == 1) || (!($masterKey =~ m/^[0-9a-fA-F]+$/))) {
	    	print(STDERR '** $masterKey ' . "must be an even number of hexadecimal digits.\n");
		$errors++;
	    } elsif (length($masterKey) < 64) {
	    	print(STDERR 'Warning: $masterKey should be at least 64 hexadecimal' . "\n");
		print(STDERR "         digits for best security.  You specified only " .
		    length($masterKey) . ".\n");
		$warnings++;
	    }
	}

	#   Reality check problem generation parameters
	if ($sentience) {
	    if ($returnTime == 0) {
	    	print(STDERR 'Warning: $returnTime set to 0.  User will not be' . "\n");
		print(STDERR "automatically to page from which feedback was sent.\n");
		$warnings++;
	    } elsif ($returnTime < 5) {
	    	print(STDERR '** $returnTime set to ' . "$returnTime seconds.  That's\n");
		print(STDERR "   probably too short for the user to read the confirmation page.\n");
		$errors++;
	    }

	    if ($minorder <= 0) {
	    	print(STDERR '** $minorder set to ' . "$minorder.  Problem is trivial or solution non-integral.\n");
		$errors++;
	    }
	    if ($maxorder <= 0) {
	    	print(STDERR '** $maxorder set to ' . "$maxorder.  Problem is trivial or solution non-integral.\n");
		$errors++;
	    }
	    if ($minorder > $maxorder) {
	    	print(STDERR '** $minorder ' . "($minorder) is greater than " . '$maxorder ' . "($maxorder)\n");
		$errors++;
	    }

	    if ($mincoeff >= $maxcoeff) {
	    	print(STDERR '** $mincoeff ' . "($mincoeff) must be less than " . '$maxcoeff ' . "($maxcoeff)\n");
		$errors++;
	    }

	    if ($minanswer  >= $maxanswer) {
	    	print(STDERR '** $minanswer ' . "($minanswer) must be less than " . '$maxanswer ' . "($maxanswer)\n");
		$errors++;
	    }

	    if ($leastanswer < 1) {
	    	print(STDERR '** $leastanswer ' . "($leastanswer) must be greater than or equal to 1.\n");
		$errors++;
	    }

    	    #	Verify worst case right hand side will fit in 32 bits
	    my $bigcoeff = (abs($mincoeff) > abs($maxcoeff)) ? abs($mincoeff) : abs($maxcoeff);
	    my $biganswer = (abs($minanswer) > abs($maxanswer)) ? abs($minanswer) : abs($maxanswer);
	    my $v = 0;
	    for (my $p = $maxorder; $p >= 0; $p--) {
	    	$v += $bigcoeff * ($biganswer ** $p);
	    }
	    if ($v > (2 ** 31)) {
	    	print(STDERR "** Largest possible right hand side value $v will not\n");
		print(STDERR "   fit in 32 bits.  Reduce order, coefficient, or answer\n");
		print(STDERR "   values to ensure it does.\n");
		$errors++;
	    }
	} else {
	    print(STDERR 'Warning: $sentience set to 0.  User will not be tested.' . "\n");
	    $warnings++;
	}

	#   Check for possible security risks
	if ($restrictReferer eq '') {
	    print(STDERR 'Warning: $restrictReferer null.  Form may be accessible from other sites.' . "\n");
	    $warnings++;
	}

	if ($cribIP ne '') {
	    print(STDERR 'Warning: $cribIP specified.  Be sure this is restricted to your test machine(s).' . "\n");
	    $warnings++;
	}

	if ($overrideEmail) {
	    print(STDERR 'Warning: $overrideEmail set.  Form requests can override the.' . "\n");
	    print(STDERR '         $email' . " address ($email) specified in this script.\n");
	    $warnings++;
	    if ($restrictEmail eq '') {
	    	print(STDERR '** $overrideEmail set and $restrictEmail null.  Form requests may' . "\n");
		print(STDERR "   direct mail to any address and may be used to send junk mail.\n");
		$errors++;
	    }
	}

	if ($restrictIP eq '') {
	    print(STDERR 'Warning: $restrictIP null.   White list requests may be sent from any site.' . "\n");
	    $warnings++;
	}

	#   Sanity check timeout values
	if ($blacklistTime < 0) {
	    print(STDERR '** $blacklistTime ' . "$blacklistTime is less than 0 seconds.\n");
	    $errors++;
	} elsif (($blacklistTime > 0) && ($blacklistTime < (5 * $minutes))) {
	    print(STDERR 'Warning: $blacklistTime of ' . "$blacklistTime seconds is less than\n");
	    print(STDERR "         5 minutes.  Are you sure you want such a short interval\n");
	    print(STDERR "         before the user can try again?\n");
	    $warnings++;
	}
	if ($expirationTime < 0) {
	    print(STDERR '** $expirationTime ' . "$expirationTime is less than 0 seconds.\n");
	    $errors++;
	} elsif (($expirationTime > 0) && ($expirationTime < (2 * $minutes))) {
	    print(STDERR 'Warning: $expirationTime of ' . "$expirationTime seconds is less than\n");
	    print(STDERR "         2 minutes.  Are you sure you want to allow the user so little\n");
	    print(STDERR "         time to solve the problem?\n");
	    $warnings++;
	}

    	#   Show error and warning summary
	if ($errors == 0) {
	    my $w = ($warnings > 0) ? " (warnings: $warnings)" : '';
    	    print(STDERR "$prog configuration tests passed$w.\n");
	} else {
	    my $w = ($warnings > 0) ? ", warnings: $warnings" : '';
	    print(STDERR "$errors error(s) found in $prog configuration tests$w.\n");
	}
	exit(($errors > 0) ? 1 : 0);
    }

    #	Generate a new $masterKey ready to paste into the program

    sub generateMasterKey {
    	my ($mv, $ms);

    	if (-r "/dev/random") {
	    open(RN, "</dev/random") || die("Cannot open /dev/random");
	    read(RN, $mv, 32);
	    close(RN);
	    print("(Generated by /dev/random)\n");
	} else {
	    srand(time() ^ ($$ + ($$ << 15)));
	    my $i = int(rand(200));
	    while ($i > 0) {
	    	rand();
		$i--;
	    }
	    for (my $n = 0; $n < 32; $n++) {
	    	$mv .= pack('C', int(rand(256)));
	    }
	    print("(Generated by built-in pseudorandom generator)\n");
	}
	for (my $j = 0; $j < length($mv); $j++) {
	    $ms .= sprintf("%02X", unpack("C", substr($mv, $j, 1)));
	}
	print("    my \$masterKey = '$ms';\n");
    	exit(0);
    }

    #	Encrypt an E-mail address given as $ARGV[1]

    sub encodeEmailAddress {

    	if ($#ARGV == 0) {
	    print(STDERR "Usage: $prog mailcode email_address\n");
	    exit(0);
	}

    	my $email = $ARGV[1];

    	my $enc = encrypt($email);
	my $cmail = escape_URL($enc);

	print("<input type=\"hidden\" name=\"j\" value=\"$cmail\" />\n");

	exit(0);
    }
